@@ -29,6 +29,7 @@ struct ws_content {
 };
 
 struct workspace {
+    int id;
     char *name;
     int name_as_int; /* -1 if name is not a decimal number */
     bool persistent;
@@ -60,6 +61,7 @@ struct private {
         size_t count;
     } ws_content;
 
+    bool strip_workspace_numbers;
     enum sort_mode sort_mode;
     tll(struct workspace) workspaces;
 
@@ -71,6 +73,22 @@ static int
 workspace_name_as_int(const char *name)
 {
     int name_as_int = 0;
+
+    /* First check for N:name pattern (set $ws1 “1:foobar”) */
+    const char *colon = strchr(name, ':');
+    if (colon != NULL) {
+        for (const char *p = name; p < colon; p++) {
+            if (!(*p >= '0' && *p < '9'))
+                return -1;
+
+            name_as_int *= 10;
+            name_as_int += *p - '0';
+        }
+
+        return name_as_int;
+    }
+
+    /* Then, if the name is a number *only* (set $ws1 1) */
     for (const char *p = name; *p != '\0'; p++) {
         if (!(*p >= '0' && *p <= '9'))
             return -1;
@@ -86,16 +104,19 @@ static bool
 workspace_from_json(const struct json_object *json, struct workspace *ws)
 {
     /* Always present */
-    struct json_object *name, *output, *focus;
-    if (!json_object_object_get_ex(json, "name", &name) ||
-        !json_object_object_get_ex(json, "output", &output) ||
-        !json_object_object_get_ex(json, "focus", &focus))
+    struct json_object *id, *name, *output;
+    if (!json_object_object_get_ex(json, "id", &id) ||
+        !json_object_object_get_ex(json, "name", &name) ||
+        !json_object_object_get_ex(json, "output", &output))
     {
-        LOG_ERR("workspace reply/event without 'name' and/or 'output', "
-                "and/or 'focus' properties");
+        LOG_ERR("workspace reply/event without 'name' and/or 'output' "
+                "properties");
         return false;
     }
 
+    /* Sway only */
+    struct json_object *focus = NULL;
+    json_object_object_get_ex(json, "focus", &focus);
 
     /* Optional */
     struct json_object *visible = NULL, *focused = NULL, *urgent = NULL;
@@ -105,18 +126,23 @@ workspace_from_json(const struct json_object *json, struct workspace *ws)
 
     const char *name_as_string = json_object_get_string(name);
 
-    const size_t node_count = json_object_array_length(focus);
+    const size_t node_count = focus != NULL
+        ? json_object_array_length(focus)
+        : 0;
+
     const bool is_empty = node_count == 0;
+    int name_as_int = workspace_name_as_int(name_as_string);
 
     *ws = (struct workspace) {
+        .id = json_object_get_int(id),
         .name = strdup(name_as_string),
-        .name_as_int = workspace_name_as_int(name_as_string),
+        .name_as_int = name_as_int,
         .persistent = false,
         .output = strdup(json_object_get_string(output)),
         .visible = json_object_get_boolean(visible),
         .focused = json_object_get_boolean(focused),
         .urgent = json_object_get_boolean(urgent),
-        .empty = is_empty,
+        .empty = is_empty && json_object_get_boolean(focused),
         .window = {.title = NULL, .pid = -1},
     };
 
@@ -199,12 +225,12 @@ workspace_add(struct private *m, struct workspace ws)
 }
 
 static void
-workspace_del(struct private *m, const char *name)
+workspace_del(struct private *m, int id)
 {
     tll_foreach(m->workspaces, it) {
         struct workspace *ws = &it->item;
 
-        if (strcmp(ws->name, name) != 0)
+        if (ws->id != id)
             continue;
 
         workspace_free(ws);
@@ -214,18 +240,18 @@ workspace_del(struct private *m, const char *name)
 }
 
 static struct workspace *
-workspace_lookup(struct private *m, const char *name)
+workspace_lookup(struct private *m, int id)
 {
     tll_foreach(m->workspaces, it) {
         struct workspace *ws = &it->item;
-        if (strcmp(ws->name, name) == 0)
+        if (ws->id == id)
             return ws;
     }
     return NULL;
 }
 
 static bool
-handle_get_version_reply(int type, const struct json_object *json, void *_m)
+handle_get_version_reply(int sock, int type, const struct json_object *json, void *_m)
 {
     struct json_object *version;
     if (!json_object_object_get_ex(json, "human_readable", &version)) {
@@ -238,7 +264,7 @@ handle_get_version_reply(int type, const struct json_object *json, void *_m)
 }
 
 static bool
-handle_subscribe_reply(int type, const struct json_object *json, void *_m)
+handle_subscribe_reply(int sock, int type, const struct json_object *json, void *_m)
 {
     struct json_object *success;
     if (!json_object_object_get_ex(json, "success", &success)) {
@@ -257,12 +283,12 @@ handle_subscribe_reply(int type, const struct json_object *json, void *_m)
 static bool
 workspace_update_or_add(struct private *m, const struct json_object *ws_json)
 {
-    struct json_object *name;
-    if (!json_object_object_get_ex(ws_json, "name", &name))
+    struct json_object *_id;
+    if (!json_object_object_get_ex(ws_json, "id", &_id))
         return false;
 
-    const char *name_as_string = json_object_get_string(name);
-    struct workspace *already_exists = workspace_lookup(m, name_as_string);
+    const int id = json_object_get_int(_id);
+    struct workspace *already_exists = workspace_lookup(m, id);
 
     if (already_exists != NULL) {
         bool persistent = already_exists->persistent;
@@ -284,7 +310,7 @@ workspace_update_or_add(struct private *m, const struct json_object *ws_json)
 }
 
 static bool
-handle_get_workspaces_reply(int type, const struct json_object *json, void *_mod)
+handle_get_workspaces_reply(int sock, int type, const struct json_object *json, void *_mod)
 {
     struct module *mod = _mod;
     struct private *m = mod->private;
@@ -311,7 +337,7 @@ err:
 }
 
 static bool
-handle_workspace_event(int type, const struct json_object *json, void *_mod)
+handle_workspace_event(int sock, int type, const struct json_object *json, void *_mod)
 {
     struct module *mod = _mod;
     struct private *m = mod->private;
@@ -327,23 +353,19 @@ handle_workspace_event(int type, const struct json_object *json, void *_mod)
     bool is_init = strcmp(change_str, "init") == 0;
     bool is_empty = strcmp(change_str, "empty") == 0;
     bool is_focused = strcmp(change_str, "focus") == 0;
+    bool is_rename = strcmp(change_str, "rename") == 0;
+    bool is_move = strcmp(change_str, "move") == 0;
     bool is_urgent = strcmp(change_str, "urgent") == 0;
-    bool is_reload = strcmp(change_str, "reload") == 0;
 
-    if (is_reload) {
-        LOG_WARN("unimplemented: 'reload' event");
-        return true;
-    }
-
-    struct json_object *current, *_current_name;
+    struct json_object *current, *_current_id;
     if (!json_object_object_get_ex(json, "current", &current) ||
-        !json_object_object_get_ex(current, "name", &_current_name))
+        !json_object_object_get_ex(current, "id", &_current_id))
     {
-        LOG_ERR("workspace event without 'current' and/or 'name' properties");
+        LOG_ERR("workspace event without 'current' and/or 'id' properties");
         return false;
     }
 
-    const char *current_name = json_object_get_string(_current_name);
+    int current_id = json_object_get_int(_current_id);
 
     mtx_lock(&mod->lock);
 
@@ -353,23 +375,22 @@ handle_workspace_event(int type, const struct json_object *json, void *_mod)
     }
 
     else if (is_empty) {
-        struct workspace *ws = workspace_lookup(m, current_name);
+        struct workspace *ws = workspace_lookup(m, current_id);
         assert(ws != NULL);
 
         if (!ws->persistent)
-            workspace_del(m, current_name);
+            workspace_del(m, current_id);
         else {
             workspace_free(ws);
-            ws->name = strdup(current_name);
             ws->empty = true;
             assert(ws->persistent);
         }
     }
 
     else if (is_focused) {
-        struct json_object *old, *_old_name, *urgent;
+        struct json_object *old, *_old_id, *urgent;
         if (!json_object_object_get_ex(json, "old", &old) ||
-            !json_object_object_get_ex(old, "name", &_old_name) ||
+            !json_object_object_get_ex(old, "id", &_old_id) ||
             !json_object_object_get_ex(current, "urgent", &urgent))
         {
             LOG_ERR("workspace 'focused' event without 'old', 'name' and/or 'urgent' property");
@@ -377,7 +398,7 @@ handle_workspace_event(int type, const struct json_object *json, void *_mod)
             return false;
         }
 
-        struct workspace *w = workspace_lookup(m, current_name);
+        struct workspace *w = workspace_lookup(m, current_id);
         assert(w != NULL);
 
         LOG_DBG("w: %s", w->name);
@@ -394,12 +415,61 @@ handle_workspace_event(int type, const struct json_object *json, void *_mod)
         w->visible = true;
 
         /* Old workspace is no longer focused */
-        const char *old_name = json_object_get_string(_old_name);
-        struct workspace *old_w = workspace_lookup(m, old_name);
+        int old_id = json_object_get_int(_old_id);
+        struct workspace *old_w = workspace_lookup(m, old_id);
         if (old_w != NULL)
             old_w->focused = false;
     }
 
+    else if (is_rename) {
+        struct workspace *w = workspace_lookup(m, current_id);
+        assert(w != NULL);
+
+        struct json_object *_current_name;
+        if (!json_object_object_get_ex(current, "name", &_current_name)) {
+            LOG_ERR("workspace 'rename' event without 'name' property");
+            mtx_unlock(&mod->lock);
+            return false;
+        }
+
+        free(w->name);
+        w->name = strdup(json_object_get_string(_current_name));
+        w->name_as_int = workspace_name_as_int(w->name);
+
+        /* Re-add the workspace to ensure correct sorting */
+        struct workspace ws = *w;
+        tll_foreach(m->workspaces, it) {
+            if (it->item.id == current_id) {
+                tll_remove(m->workspaces, it);
+                break;
+            }
+        }
+        workspace_add(m, ws);
+    }
+
+    else if (is_move) {
+        struct workspace *w = workspace_lookup(m, current_id);
+        assert(w != NULL);
+
+        struct json_object *_current_output;
+        if (!json_object_object_get_ex(current, "output", &_current_output)) {
+            LOG_ERR("workspace 'move' event without 'output' property");
+            mtx_unlock(&mod->lock);
+            return false;
+        }
+
+        free(w->output);
+        w->output = strdup(json_object_get_string(_current_output));
+
+        /*
+         * If the moved workspace was focused, schedule a full update because
+         * visibility for other workspaces may have changed.
+         */
+        if (w->focused) {
+            i3_send_pkg(sock, I3_IPC_MESSAGE_TYPE_GET_WORKSPACES, NULL);
+        }
+    }
+
     else if (is_urgent) {
         struct json_object *urgent;
         if (!json_object_object_get_ex(current, "urgent", &urgent)) {
@@ -408,10 +478,14 @@ handle_workspace_event(int type, const struct json_object *json, void *_mod)
             return false;
         }
 
-        struct workspace *w = workspace_lookup(m, current_name);
+        struct workspace *w = workspace_lookup(m, current_id);
         w->urgent = json_object_get_boolean(urgent);
     }
 
+    else {
+        LOG_WARN("unimplemented workspace event '%s'", change_str);
+    }
+
     m->dirty = true;
     mtx_unlock(&mod->lock);
     return true;
@@ -422,7 +496,7 @@ err:
 }
 
 static bool
-handle_window_event(int type, const struct json_object *json, void *_mod)
+handle_window_event(int sock, int type, const struct json_object *json, void *_mod)
 {
     struct module *mod = _mod;
     struct private *m = mod->private;
@@ -556,7 +630,7 @@ handle_window_event(int type, const struct json_object *json, void *_mod)
 }
 
 static bool
-handle_mode_event(int type, const struct json_object *json, void *_mod)
+handle_mode_event(int sock, int type, const struct json_object *json, void *_mod)
 {
     struct module *mod = _mod;
     struct private *m = mod->private;
@@ -598,7 +672,7 @@ run(struct module *mod)
     if (!i3_get_socket_address(&addr))
         return 1;
 
-    int sock = socket(AF_UNIX, SOCK_STREAM, 0);
+    int sock = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0);
     if (sock == -1) {
         LOG_ERRNO("failed to create UNIX socket");
         return 1;
@@ -615,9 +689,16 @@ run(struct module *mod)
     for (size_t i = 0; i < m->persistent_count; i++) {
         const char *name_as_string = m->persistent_workspaces[i];
 
+        int name_as_int = workspace_name_as_int(name_as_string);
+        if (m->strip_workspace_numbers) {
+            const char *colon = strchr(name_as_string, ':');
+            if (colon != NULL)
+                name_as_string = colon++;
+        }
+
         struct workspace ws = {
             .name = strdup(name_as_string),
-            .name_as_int = workspace_name_as_int(name_as_string),
+            .name_as_int = name_as_int,
             .persistent = true,
             .empty = true,
         };
@@ -679,7 +760,7 @@ ws_content_for_name(struct private *m, const char *name)
 }
 
 static const char *
-description(struct module *mod)
+description(const struct module *mod)
 {
     return "i3/sway";
 }
@@ -725,9 +806,17 @@ content(struct module *mod)
                 ws->window.title,
                 m->mode);
 
+        const char *name = ws->name;
+
+        if (m->strip_workspace_numbers) {
+            const char *colon = strchr(name, ':');
+            if (colon != NULL)
+                name = colon + 1;
+        }
+
         struct tag_set tags = {
             .tags = (struct tag *[]){
-                tag_new_string(mod, "name", ws->name),
+                tag_new_string(mod, "name", name),
                 tag_new_bool(mod, "visible", ws->visible),
                 tag_new_bool(mod, "focused", ws->focused),
                 tag_new_bool(mod, "urgent", ws->urgent),
@@ -778,7 +867,8 @@ static struct module *
 i3_new(struct i3_workspaces workspaces[], size_t workspace_count,
        int left_spacing, int right_spacing, enum sort_mode sort_mode,
        size_t persistent_count,
-       const char *persistent_workspaces[static persistent_count])
+       const char *persistent_workspaces[static persistent_count],
+       bool strip_workspace_numbers)
 {
     struct private *m = calloc(1, sizeof(*m));
 
@@ -794,6 +884,7 @@ i3_new(struct i3_workspaces workspaces[], size_t workspace_count,
         m->ws_content.v[i].content = workspaces[i].content;
     }
 
+    m->strip_workspace_numbers = strip_workspace_numbers;
     m->sort_mode = sort_mode;
 
     m->persistent_count = persistent_count;
@@ -821,6 +912,8 @@ from_conf(const struct yml_node *node, struct conf_inherit inherited)
     const struct yml_node *right_spacing = yml_get_value(node, "right-spacing");
     const struct yml_node *sort = yml_get_value(node, "sort");
     const struct yml_node *persistent = yml_get_value(node, "persistent");
+    const struct yml_node *strip_workspace_number = yml_get_value(
+        node, "strip-workspace-numbers");
 
     int left = spacing != NULL ? yml_value_as_int(spacing) :
         left_spacing != NULL ? yml_value_as_int(left_spacing) : 0;
@@ -859,7 +952,9 @@ from_conf(const struct yml_node *node, struct conf_inherit inherited)
     }
 
     return i3_new(workspaces, yml_dict_length(c), left, right, sort_mode,
-                  persistent_count, persistent_workspaces);
+                  persistent_count, persistent_workspaces,
+                  (strip_workspace_number != NULL
+                   ? yml_value_as_bool(strip_workspace_number) : false));
 }
 
 static bool
@@ -914,6 +1009,7 @@ verify_conf(keychain_t *chain, const struct yml_node *node)
         {"right-spacing", false, &conf_verify_unsigned},
         {"sort", false, &verify_sort},
         {"persistent", false, &verify_persistent},
+        {"strip-workspace-numbers", false, &conf_verify_bool},
         {"content", true, &verify_content},
         {"anchors", false, NULL},
         {NULL, false, NULL},
