@@ -11,6 +11,9 @@
 #include <poll.h>
 #include <sys/timerfd.h>
 
+#include <sys/stat.h>
+#include <fcntl.h>
+
 #include <arpa/inet.h>
 #include <linux/if.h>
 #include <linux/netlink.h>
@@ -31,6 +34,11 @@
 
 #define UNUSED __attribute__((unused))
 
+struct rt_stats_msg {
+    struct rtmsg rth;
+    struct rtnl_link_stats64 stats;
+};
+
 struct af_addr {
     int family;
     union {
@@ -46,11 +54,13 @@ struct private {
 
     int genl_sock;
     int rt_sock;
+    int urandom_fd;
 
     struct {
         uint16_t family_id;
         uint32_t get_interface_seq_nr;
         uint32_t get_station_seq_nr;
+        uint32_t get_scan_seq_nr;
     } nl80211;
 
     bool get_addresses;
@@ -68,6 +78,12 @@ struct private {
     int signal_strength_dbm;
     uint32_t rx_bitrate;
     uint32_t tx_bitrate;
+
+    uint64_t ul_speed;
+    uint64_t ul_bits;
+
+    uint64_t dl_speed;
+    uint64_t dl_bits;
 };
 
 static void
@@ -79,6 +95,9 @@ destroy(struct module *mod)
 
     m->label->destroy(m->label);
 
+    if (m->urandom_fd >= 0)
+        close(m->urandom_fd);
+
     tll_free(m->addrs);
     free(m->ssid);
     free(m->iface);
@@ -88,10 +107,10 @@ destroy(struct module *mod)
 }
 
 static const char *
-description(struct module *mod)
+description(const struct module *mod)
 {
     static char desc[32];
-    struct private *m = mod->private;
+    const struct private *m = mod->private;
 
     snprintf(desc, sizeof(desc), "net(%s)", m->iface);
     return desc;
@@ -143,10 +162,12 @@ content(struct module *mod)
             tag_new_string(mod, "ipv6", ipv6_str),
             tag_new_string(mod, "ssid", m->ssid),
             tag_new_int(mod, "signal", m->signal_strength_dbm),
-            tag_new_int(mod, "rx-bitrate", m->rx_bitrate / 1000 / 1000),
-            tag_new_int(mod, "tx-bitrate", m->tx_bitrate / 1000 / 1000),
+            tag_new_int(mod, "rx-bitrate", m->rx_bitrate),
+            tag_new_int(mod, "tx-bitrate", m->tx_bitrate),
+            tag_new_float(mod, "dl-speed", m->dl_speed),
+            tag_new_float(mod, "ul-speed", m->ul_speed),
         },
-        .count = 11,
+        .count = 13,
     };
 
     mtx_unlock(&mod->lock);
@@ -167,7 +188,7 @@ nl_pid_value(void)
 static int
 netlink_connect_rt(void)
 {
-    int sock = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
+    int sock = socket(AF_NETLINK, SOCK_RAW | SOCK_CLOEXEC, NETLINK_ROUTE);
     if (sock == -1) {
         LOG_ERRNO("failed to create netlink socket");
         return -1;
@@ -191,7 +212,7 @@ netlink_connect_rt(void)
 static int
 netlink_connect_genl(void)
 {
-    int sock = socket(AF_NETLINK, SOCK_RAW, NETLINK_GENERIC);
+    int sock = socket(AF_NETLINK, SOCK_RAW | SOCK_CLOEXEC, NETLINK_GENERIC);
     if (sock == -1) {
         LOG_ERRNO("failed to create netlink socket");
         return -1;
@@ -252,6 +273,36 @@ send_rt_request(struct private *m, int request)
     return true;
 }
 
+static bool
+send_rt_getstats_request(struct private *m)
+{
+    struct {
+        struct nlmsghdr hdr;
+        struct if_stats_msg rt;
+    } req = {
+        .hdr = {
+            .nlmsg_len = NLMSG_LENGTH(sizeof(req.rt)),
+            .nlmsg_type = RTM_GETSTATS,
+            .nlmsg_flags = NLM_F_REQUEST,
+            .nlmsg_seq = 1,
+            .nlmsg_pid = nl_pid_value(),
+        },
+
+        .rt = {
+            .ifindex = m->ifindex,
+            .filter_mask = IFLA_STATS_LINK_64,
+            .family = AF_UNSPEC,
+        },
+    };
+
+    if (!send_nlmsg(m->rt_sock, &req, req.hdr.nlmsg_len)) {
+        LOG_ERRNO("%s: failed to send netlink RT getstats request (%d)",
+                m->iface, RTM_GETSTATS);
+        return false;
+    }
+    return true;
+}
+
 static bool
 send_ctrl_get_family_request(struct private *m)
 {
@@ -368,7 +419,12 @@ send_nl80211_get_interface(struct private *m)
 
     LOG_DBG("%s: sending nl80211 get-interface request", m->iface);
 
-    uint32_t seq = time(NULL);
+    uint32_t seq;
+    if (read(m->urandom_fd, &seq, sizeof(seq)) != sizeof(seq)) {
+        LOG_ERRNO("failed to read from /dev/urandom");
+        return false;
+    }
+
     if (send_nl80211_request(m, NL80211_CMD_GET_INTERFACE, NLM_F_REQUEST, seq)) {
         m->nl80211.get_interface_seq_nr = seq;
         return true;
@@ -387,7 +443,12 @@ send_nl80211_get_station(struct private *m)
 
     LOG_DBG("%s: sending nl80211 get-station request", m->iface);
 
-    uint32_t seq = time(NULL);
+    uint32_t seq;
+    if (read(m->urandom_fd, &seq, sizeof(seq)) != sizeof(seq)) {
+        LOG_ERRNO("failed to read from /dev/urandom");
+        return false;
+    }
+
     if (send_nl80211_request(
             m, NL80211_CMD_GET_STATION, NLM_F_REQUEST | NLM_F_DUMP, seq))
     {
@@ -397,6 +458,32 @@ send_nl80211_get_station(struct private *m)
         return false;
 }
 
+static bool
+send_nl80211_get_scan(struct private *m)
+{
+    if (m->nl80211.get_scan_seq_nr > 0) {
+        LOG_ERR(
+            "%s: nl80211 get-scan request already in progress", m->iface);
+        return true;
+    }
+
+    LOG_DBG("%s: sending nl80211 get-scan request", m->iface);
+
+    uint32_t seq;
+    if (read(m->urandom_fd, &seq, sizeof(seq)) != sizeof(seq)) {
+        LOG_ERRNO("failed to read from /dev/urandom");
+        return false;
+    }
+
+    if (send_nl80211_request(
+            m, NL80211_CMD_GET_SCAN, NLM_F_REQUEST | NLM_F_DUMP, seq))
+    {
+        m->nl80211.get_scan_seq_nr = seq;
+        return true;
+    } else
+        return false;
+}
+
 static bool
 find_my_ifindex(struct module *mod, const struct ifinfomsg *msg, size_t len)
 {
@@ -751,6 +838,9 @@ handle_nl80211_new_interface(struct module *mod, uint16_t type, bool nested,
     case NL80211_ATTR_SSID: {
         const char *ssid = payload;
 
+        if (m->ssid == NULL || strncmp(m->ssid, ssid, len) != 0)
+            LOG_INFO("%s: SSID: %.*s", m->iface, (int)len, ssid);
+
         mtx_lock(&mod->lock);
         free(m->ssid);
         m->ssid = strndup(ssid, len);
@@ -891,6 +981,103 @@ handle_nl80211_new_station(struct module *mod, uint16_t type, bool nested,
     return true;
 }
 
+static bool
+handle_ies(struct module *mod, const void *_ies, size_t len)
+{
+    struct private *m = mod->private;
+    const uint8_t *ies = _ies;
+
+    while (len >= 2 && len - 2 >= ies[1]) {
+        switch (ies[0]) {
+        case 0: {  /* SSID */
+            const char *ssid = (const char *)&ies[2];
+            const size_t ssid_len = ies[1];
+
+            if (m->ssid == NULL || strncmp(m->ssid, ssid, ssid_len) != 0)
+                LOG_INFO("%s: SSID: %.*s", m->iface, (int)ssid_len, ssid);
+
+            mtx_lock(&mod->lock);
+            free(m->ssid);
+            m->ssid = strndup(ssid, ssid_len);
+            mtx_unlock(&mod->lock);
+
+            mod->bar->refresh(mod->bar);
+            break;
+        }
+        }
+        len -= ies[1] + 2;
+        ies += ies[1] + 2;
+    }
+
+    return true;
+}
+
+struct scan_results_context {
+    bool associated;
+
+    const void *ies;
+    size_t ies_size;
+};
+
+static bool
+handle_nl80211_bss(struct module *mod, uint16_t type, bool nested,
+                   const void *payload, size_t len, void *_ctx)
+{
+    struct private *m UNUSED = mod->private;
+    struct scan_results_context *ctx = _ctx;
+
+    switch (type) {
+    case NL80211_BSS_STATUS: {
+        const uint32_t status = *(uint32_t *)payload;
+
+        if (status == NL80211_BSS_STATUS_ASSOCIATED) {
+            ctx->associated = true;
+
+            if (ctx->ies != NULL) {
+                /* Deferred handling of BSS_INFORMATION_ELEMENTS */
+                return handle_ies(mod, ctx->ies, ctx->ies_size);
+            }
+        }
+        break;
+    }
+
+    case NL80211_BSS_INFORMATION_ELEMENTS:
+        if (ctx->associated)
+            return handle_ies(mod, payload, len);
+        else {
+            /*
+             * We’re either not associated, or, we haven’t seen the
+             * BSS_STATUS attribute yet.
+             *
+             * Save a pointer to the IES payload, so that we can
+             * process it later, if we see a
+             *   BSS_STATUS == BSS_STATUS_ASSOCIATED.
+             */
+            ctx->ies = payload;
+            ctx->ies_size = len;
+        }
+    }
+
+    return true;
+}
+
+static bool
+handle_nl80211_scan_results(struct module *mod, uint16_t type, bool nested,
+                            const void *payload, size_t len)
+{
+    struct private *m UNUSED = mod->private;
+
+    struct scan_results_context ctx = {0};
+
+    switch (type) {
+    case NL80211_ATTR_BSS:
+        foreach_nlattr_nested(mod, payload, len, &handle_nl80211_bss, &ctx);
+        break;
+    }
+
+    return true;
+}
+
 /*
  * Reads at least one (possibly more) message.
  *
@@ -929,6 +1116,23 @@ netlink_receive_messages(int sock, void **reply, size_t *len)
     return true;
 }
 
+static void
+handle_stats(struct module *mod, struct rt_stats_msg *msg)
+{
+    struct private *m = mod->private;
+    uint64_t ul_bits = msg->stats.tx_bytes*8;
+    uint64_t dl_bits = msg->stats.rx_bytes*8;
+
+    if (m->ul_bits != 0) {
+        m->ul_speed = (ul_bits - m->ul_bits) / m->poll_interval;
+    }
+    if (m->dl_bits != 0) {
+        m->dl_speed = (dl_bits - m->dl_bits) / m->poll_interval;
+    }
+    m->ul_bits = ul_bits;
+    m->dl_bits = dl_bits;
+}
+
 static bool
 parse_rt_reply(struct module *mod, const struct nlmsghdr *hdr, size_t len)
 {
@@ -967,6 +1171,11 @@ parse_rt_reply(struct module *mod, const struct nlmsghdr *hdr, size_t len)
             handle_address(mod, hdr->nlmsg_type, msg, msg_len);
             break;
         }
+        case RTM_NEWSTATS: {
+            struct rt_stats_msg *msg = NLMSG_DATA(hdr);
+            handle_stats(mod, msg);
+            break;
+        }
 
         case NLMSG_ERROR:{
             const struct nlmsgerr *err = NLMSG_DATA(hdr);
@@ -1001,6 +1210,11 @@ parse_genl_reply(struct module *mod, const struct nlmsghdr *hdr, size_t len)
                 /* Current request is now considered complete */
                 m->nl80211.get_station_seq_nr = 0;
             }
+
+            else if (hdr->nlmsg_seq == m->nl80211.get_scan_seq_nr) {
+                /* Current request is now considered complete */
+                m->nl80211.get_scan_seq_nr = 0;
+            }
         }
 
         else if (hdr->nlmsg_type == GENL_ID_CTRL) {
@@ -1019,8 +1233,6 @@ parse_genl_reply(struct module *mod, const struct nlmsghdr *hdr, size_t len)
                     LOG_DBG("%s: got interface information", m->iface);
                     foreach_nlattr(
                         mod, genl, msg_size, &handle_nl80211_new_interface);
-
-                    LOG_INFO("%s: SSID: %s", m->iface, m->ssid);
                 }
                 break;
 
@@ -1062,10 +1274,22 @@ parse_genl_reply(struct module *mod, const struct nlmsghdr *hdr, size_t len)
                     foreach_nlattr(mod, genl, msg_size, &handle_nl80211_new_station);
                 }
 
-                LOG_INFO("%s: signal: %d dBm, RX=%u Mbit/s, TX=%u Mbit/s",
-                         m->iface, m->signal_strength_dbm,
-                         m->rx_bitrate / 1000 / 1000,
-                         m->tx_bitrate / 1000 / 1000);
+                LOG_DBG("%s: signal: %d dBm, RX=%u Mbit/s, TX=%u Mbit/s",
+                        m->iface, m->signal_strength_dbm,
+                        m->rx_bitrate / 1000 / 1000,
+                        m->tx_bitrate / 1000 / 1000);
+
+                /* Can’t issue both get-station and get-scan at the
+                 * same time. So, always run a get-scan when a
+                 * get-station is complete */
+                send_nl80211_get_scan(m);
+                break;
+
+            case NL80211_CMD_NEW_SCAN_RESULTS:
+                if (nl80211_is_for_us(mod, genl, msg_size)) {
+                    LOG_DBG("%s: got scan results", m->iface);
+                    foreach_nlattr(mod, genl, msg_size, &handle_nl80211_scan_results);
+                }
                 break;
 
             default:
@@ -1083,7 +1307,8 @@ parse_genl_reply(struct module *mod, const struct nlmsghdr *hdr, size_t len)
             else if (nl_errno == ENOENT)
                 ; /* iface down? */
             else
-                LOG_ERRNO_P(nl_errno, "%s: nl80211 reply", m->iface);
+                LOG_ERRNO_P(nl_errno, "%s: nl80211 reply (seq-nr: %u)",
+                            m->iface, hdr->nlmsg_seq);
         }
 
         else {
@@ -1200,6 +1425,7 @@ run(struct module *mod)
             }
 
             send_nl80211_get_station(m);
+            send_rt_getstats_request(m);
         }
     }
 
@@ -1219,6 +1445,12 @@ run(struct module *mod)
 static struct module *
 network_new(const char *iface, struct particle *label, int poll_interval)
 {
+    int urandom_fd = open("/dev/urandom", O_RDONLY);
+    if (urandom_fd < 0) {
+        LOG_ERRNO("failed to open /dev/urandom");
+        return NULL;
+    }
+
     struct private *priv = calloc(1, sizeof(*priv));
     priv->iface = strdup(iface);
     priv->label = label;
@@ -1226,6 +1458,7 @@ network_new(const char *iface, struct particle *label, int poll_interval)
 
     priv->genl_sock = -1;
     priv->rt_sock = -1;
+    priv->urandom_fd = urandom_fd;
     priv->nl80211.family_id = -1;
     priv->get_addresses = true;
     priv->ifindex = -1;
