@@ -16,6 +16,8 @@
 #include "river-status-unstable-v1.h"
 #include "xdg-output-unstable-v1.h"
 
+#define min(x, y) ((x) < (y) ? (x) : (y))
+
 struct private;
 
 struct output {
@@ -39,6 +41,7 @@ struct seat {
     uint32_t wl_name;
     char *name;
 
+    char *mode;
     char *title;
     struct output *output;
 };
@@ -51,7 +54,6 @@ struct private {
     struct particle *title;
     bool all_monitors;
 
-    bool is_starting_up;
     tll(struct output) outputs;
     tll(struct seat) seats;
 };
@@ -68,7 +70,7 @@ destroy(struct module *mod)
 }
 
 static const char *
-description(struct module *mod)
+description(const struct module *mod)
 {
     return "river";
 }
@@ -109,8 +111,7 @@ content(struct module *mod)
         }
     }
 
-    const size_t seat_count = m->title != NULL && !m->is_starting_up
-        ? tll_length(m->seats) : 0;
+    const size_t seat_count = m->title != NULL ? tll_length(m->seats) : 0;
     struct exposable *tag_parts[32 + seat_count];
 
     for (unsigned i = 0; i < 32; i++) {
@@ -130,7 +131,7 @@ content(struct module *mod)
 
 #if 0
         LOG_DBG("tag: #%u, visible=%d, focused=%d, occupied=%d, state=%s",
-                i, visible, focused, occupied & (1u << i), state);
+                i, is_visible, is_focused, is_occupied & (1u << i), state);
 #endif
 
         struct tag_set tags = {
@@ -149,7 +150,7 @@ content(struct module *mod)
         tag_set_destroy(&tags);
     }
 
-    if (m->title != NULL && !m->is_starting_up) {
+    if (m->title != NULL) {
         size_t i = 32;
         tll_foreach(m->seats, it) {
             const struct seat *seat = &it->item;
@@ -158,8 +159,9 @@ content(struct module *mod)
                 .tags = (struct tag *[]){
                     tag_new_string(mod, "seat", seat->name),
                     tag_new_string(mod, "title", seat->title),
+                    tag_new_string(mod, "mode", seat->mode),
                 },
-                .count = 2,
+                .count = 3,
             };
 
             tag_parts[i++] = m->title->instantiate(m->title, &tags);
@@ -185,6 +187,11 @@ verify_iface_version(const char *iface, uint32_t version, uint32_t wanted)
 static void
 output_destroy(struct output *output)
 {
+    tll_foreach(output->m->seats, it) {
+        struct seat *seat = &it->item;
+        if (seat->output == output)
+            seat->output = NULL;
+    }
     free(output->name);
     if (output->status != NULL)
         zriver_output_status_v1_destroy(output->status);
@@ -199,6 +206,7 @@ seat_destroy(struct seat *seat)
 {
     free(seat->title);
     free(seat->name);
+    free(seat->mode);
     if (seat->status != NULL)
         zriver_seat_status_v1_destroy(seat->status);
     if (seat->wl_seat != NULL)
@@ -318,14 +326,20 @@ static struct zxdg_output_v1_listener xdg_output_listener = {
 };
 
 static void
-instantiate_output(struct output *output)
+update_output(struct output *output)
 {
-    if (output->m->is_starting_up)
-        return;
-
     assert(output->wl_output != NULL);
 
-    if (output->m->status_manager != NULL && output->status == NULL) {
+    if (output->m->status_manager != NULL) {
+        /*
+         * Bind river output status, if we have already bound the status manager
+         */
+
+        if (output->status != NULL) {
+            zriver_output_status_v1_destroy(output->status);
+            output->status = NULL;
+        }
+
         output->status = zriver_status_manager_v1_get_river_output_status(
             output->m->status_manager, output->wl_output);
 
@@ -377,7 +391,7 @@ focused_output(void *data, struct zriver_seat_status_v1 *zriver_seat_status_v1,
 
 static void
 unfocused_output(void *data, struct zriver_seat_status_v1 *zriver_seat_status_v1,
-				 struct wl_output *wl_output)
+                 struct wl_output *wl_output)
 {
     struct seat *seat = data;
     struct private *m = seat->m;
@@ -435,10 +449,34 @@ focused_view(void *data, struct zriver_seat_status_v1 *zriver_seat_status_v1,
     }
 }
 
+#if defined(ZRIVER_SEAT_STATUS_V1_MODE_SINCE_VERSION)
+static void
+mode(void *data, struct zriver_seat_status_v1 *zriver_seat_status_v1,
+     const char *name)
+{
+    struct seat *seat = data;
+    struct module *mod = seat->m->mod;
+
+    mtx_lock(&mod->lock);
+    {
+        free(seat->mode);
+        seat->mode = strdup(name);
+        mtx_unlock(&mod->lock);
+    }
+    mod->bar->refresh(mod->bar);
+
+    LOG_DBG("seat: %s, current mode: %s", seat->name, seat->mode);
+}
+
+#endif
+
 static const struct zriver_seat_status_v1_listener river_seat_status_listener = {
     .focused_output = &focused_output,
     .unfocused_output = &unfocused_output,
     .focused_view = &focused_view,
+#if defined(ZRIVER_SEAT_STATUS_V1_MODE_SINCE_VERSION)
+    .mode = &mode,
+#endif
 };
 
 static void
@@ -468,16 +506,18 @@ static const struct wl_seat_listener seat_listener = {
 };
 
 static void
-instantiate_seat(struct seat *seat)
+update_seat(struct seat *seat)
 {
     assert(seat->wl_seat != NULL);
 
-    if (seat->m->is_starting_up)
-        return;
-
     if (seat->m->status_manager == NULL)
         return;
 
+    if (seat->status != NULL) {
+        zriver_seat_status_v1_destroy(seat->status);
+        seat->status = NULL;
+    }
+
     seat->status = zriver_status_manager_v1_get_river_seat_status(
         seat->m->status_manager, seat->wl_seat);
 
@@ -513,7 +553,9 @@ handle_global(void *data, struct wl_registry *registry,
 
         mtx_lock(&m->mod->lock);
         tll_push_back(m->outputs, output);
-        instantiate_output(&tll_back(m->outputs));
+        update_output(&tll_back(m->outputs));
+        tll_foreach(m->seats, it)
+            update_seat(&it->item);
         mtx_unlock(&m->mod->lock);
     }
 
@@ -527,7 +569,7 @@ handle_global(void *data, struct wl_registry *registry,
 
         mtx_lock(&m->mod->lock);
         tll_foreach(m->outputs, it)
-            instantiate_output(&it->item);
+            update_output(&it->item);
         mtx_unlock(&m->mod->lock);
     }
 
@@ -547,7 +589,7 @@ handle_global(void *data, struct wl_registry *registry,
         struct seat *seat = &tll_back(m->seats);
 
         wl_seat_add_listener(wl_seat, &seat_listener, seat);
-        instantiate_seat(seat);
+        update_seat(seat);
         mtx_unlock(&m->mod->lock);
     }
 
@@ -557,13 +599,13 @@ handle_global(void *data, struct wl_registry *registry,
             return;
 
         m->status_manager = wl_registry_bind(
-            registry, name, &zriver_status_manager_v1_interface, required);
+            registry, name, &zriver_status_manager_v1_interface, min(version, 3));
 
         mtx_lock(&m->mod->lock);
         tll_foreach(m->outputs, it)
-            instantiate_output(&it->item);
+            update_output(&it->item);
         tll_foreach(m->seats, it)
-            instantiate_seat(&it->item);
+            update_seat(&it->item);
         mtx_unlock(&m->mod->lock);
     }
 }
@@ -630,19 +672,6 @@ run(struct module *mod)
 
     wl_display_roundtrip(display);
 
-    bool unlock_at_exit = true;
-    mtx_lock(&mod->lock);
-
-    m->is_starting_up = false;
-
-    tll_foreach(m->outputs, it)
-        instantiate_output(&it->item);
-    tll_foreach(m->seats, it)
-        instantiate_seat(&it->item);
-
-    unlock_at_exit = false;
-    mtx_unlock(&mod->lock);
-
     while (true) {
         wl_display_flush(display);
 
@@ -689,9 +718,6 @@ out:
         wl_registry_destroy(registry);
     if (display != NULL)
         wl_display_disconnect(display);
-
-    if (unlock_at_exit)
-        mtx_unlock(&mod->lock);
     return ret;
 }
 
@@ -702,7 +728,6 @@ river_new(struct particle *template, struct particle *title, bool all_monitors)
     m->template = template;
     m->title = title;
     m->all_monitors = all_monitors;
-    m->is_starting_up = true;
 
     struct module *mod = module_common_new();
     mod->private = m;
