@@ -5,6 +5,7 @@
 #include <assert.h>
 #include <unistd.h>
 #include <libgen.h>
+#include <signal.h>
 
 #include <poll.h>
 #include <fcntl.h>
@@ -12,7 +13,6 @@
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/wait.h>
-#include <sys/timerfd.h>
 
 #define LOG_MODULE "script"
 #define LOG_ENABLE_DBG 0
@@ -60,10 +60,10 @@ destroy(struct module *mod)
 }
 
 static const char *
-description(struct module *mod)
+description(const struct module *mod)
 {
     static char desc[32];
-    struct private *m = mod->private;
+    const struct private *m = mod->private;
 
     char *path = strdup(m->path);
     snprintf(desc, sizeof(desc), "script(%s)", basename(path));
@@ -313,21 +313,23 @@ data_received(struct module *mod, const char *data, size_t len)
     memcpy(&m->recv_buf.data[m->recv_buf.idx], data, len);
     m->recv_buf.idx += len;
 
-    const char *eot = memmem(m->recv_buf.data, m->recv_buf.idx, "\n\n", 2);
-    if (eot == NULL) {
-        /* End of transaction not yet available */
-        return true;
+    while (true) {
+        const char *eot = memmem(m->recv_buf.data, m->recv_buf.idx, "\n\n", 2);
+        if (eot == NULL) {
+            /* End of transaction not yet available */
+            return true;
+        }
+
+        const size_t transaction_size = eot - m->recv_buf.data + 1;
+        process_transaction(mod, transaction_size);
+
+        assert(m->recv_buf.idx >= transaction_size + 1);
+        memmove(m->recv_buf.data,
+                &m->recv_buf.data[transaction_size + 1],
+                m->recv_buf.idx - (transaction_size + 1));
+        m->recv_buf.idx -= transaction_size + 1;
     }
 
-    const size_t transaction_size = eot - m->recv_buf.data + 1;
-    process_transaction(mod, transaction_size);
-
-    assert(m->recv_buf.idx >= transaction_size + 1);
-    memmove(m->recv_buf.data,
-            &m->recv_buf.data[transaction_size + 1],
-            m->recv_buf.idx - (transaction_size + 1));
-    m->recv_buf.idx -= transaction_size + 1;
-
     return true;
 }
 
@@ -396,7 +398,7 @@ execute_script(struct module *mod)
 
     /* Stdout redirection pipe */
     int comm_pipe[2];
-    if (pipe(comm_pipe) < 0) {
+    if (pipe2(comm_pipe, O_CLOEXEC) < 0) {
         LOG_ERRNO("failed to create stdin/stdout redirection pipe");
         close(exec_pipe[0]);
         close(exec_pipe[1]);
@@ -444,7 +446,7 @@ execute_script(struct module *mod)
         close(comm_pipe[0]);
 
         /* Re-direct stdin/stdout */
-        int dev_null = open("/dev/null", O_RDONLY);
+        int dev_null = open("/dev/null", O_RDONLY | O_CLOEXEC);
         if (dev_null < 0)
             goto fail;
 
@@ -458,16 +460,6 @@ execute_script(struct module *mod)
         close(comm_pipe[1]);
         comm_pipe[1] = -1;
 
-        /* Close *all* other FDs */
-        for (int i = STDERR_FILENO + 1; i < 65536; i++) {
-            if (i == exec_pipe[1]) {
-                /* Needed for error reporting. Automatically closed
-                 * when execvp() succeeds */
-                continue;
-            }
-            close(i);
-        }
-
         execvp(m->path, argv);
 
     fail:
